--Charge abilities
local SKIN = "jetmetalsonic"

freeslot("SPR2_CHRG","S_PLAY_CHARGE","S_PLAY_CHARGETHOK","S_PLAY_CHARGEJUMP")

states[S_PLAY_CHARGE] = {SPR_PLAY, SPR2_CHRG|A, 1, nil, 0, 0, S_PLAY_CHARGE}
states[S_PLAY_CHARGETHOK] = {SPR_PLAY, SPR2_DASH|A, 4*TICRATE/5, nil, 0, 0, S_PLAY_ROLL}
states[S_PLAY_CHARGEJUMP] = {SPR_PLAY, SPR2_SPNG|A, 14, nil, 0, 0, S_PLAY_ROLL}

local MAXENERGY = 100*FRACUNIT				--Maximum available energy
local MAXCHARGE = MAXENERGY/2				--Maximum possible charge
local MINCHARGE = MAXCHARGE/5				--Minimum possible charge
local MAXCHARGETIME = 2*TICRATE				--Amount of time charging can be done
local CHRGSPD = MAXCHARGE/(17*TICRATE/15)	--Amount of charge gained per tic

local function StopCharge(p)
	S_StopSoundByID(p.mo, sfx_cdfm44)
	if p.prevcolor
		p.mo.color = p.prevcolor
		p.prevcolor = nil
	end
	p.mo.colorized = false

	local chrg = p.chargeability
	chrg.charging = false
	chrg.charge = 0
	chrg.chargetime = 0
end

local function ConvertCharge(p)
	local chrg = p.chargeability
	chrg.energy = $ - chrg.charge
	chrg.chargemod = FRACUNIT + FixedMul(5*FRACUNIT/4, FixedDiv(chrg.charge, MAXCHARGE))
	S_StartSound(p.mo, sfx_beflap)
end

local function doChargeThok(p)
	ConvertCharge(p)
	local mo = p.mo
	local chrg = p.chargeability

	local THOKSPEED = 2*chrg.storespd/3 + FixedMul(FixedMul(20*FRACUNIT, chrg.chargemod), mo.scale)
	local WATERFACTOR = FRACUNIT
	if mo.eflags & MFE_UNDERWATER
		WATERFACTOR = 3*$/5
	end
	local SUPERFACTOR = FRACUNIT
	if p.powers[pw_super]
		SUPERFACTOR = 6*$/5
	end
	P_InstaThrust(mo, mo.angle, FixedMul(FixedMul(FixedMul(THOKSPEED, WATERFACTOR), SUPERFACTOR), mo.scale))
	P_SetObjectMomZ(mo, FixedMul(5*FRACUNIT, WATERFACTOR), false)

	chrg.wallbounce = 4*TICRATE/5
	chrg.bouncestrength = chrg.chargemod

	mo.state = S_PLAY_CHARGETHOK
	SpawnAuraEffect(p)
	SpawnLaunchEffect(p, 4*FRACUNIT/5, 1)
end

local function doChargeJump(p)
	ConvertCharge(p)
	local mo = p.mo
	local chrg = p.chargeability

	local JUMPHEIGHT = FixedMul(FixedMul(15*FRACUNIT/2, chrg.chargemod), mo.scale)
	local WATERFACTOR = FRACUNIT
	if mo.eflags & MFE_UNDERWATER
		WATERFACTOR = 3*$/5
	end
	local SUPERFACTOR = FRACUNIT
	if p.powers[pw_super]
		SUPERFACTOR = 6*$/5
	end
	P_SetObjectMomZ(mo, FixedMul(FixedMul(JUMPHEIGHT, WATERFACTOR), SUPERFACTOR), false)
	P_InstaThrust(mo, mo.angle, FixedMul(chrg.storespd/3, mo.scale))

	mo.state = S_PLAY_CHARGEJUMP
	SpawnAuraEffect(p)
	SpawnLaunchEffect(p, 4*FRACUNIT/5, 0)
end

local function doChargeStomp(p)
	ConvertCharge(p)
	local mo = p.mo
	local chrg = p.chargeability

	P_InstaThrust(mo, mo.angle, FixedMul(chrg.storespd/2, mo.scale))

	chrg.stomp = true
	chrg.floorbounce = true
	chrg.bouncestrength = chrg.chargemod

	mo.state = S_PLAY_ROLL
	SpawnAuraEffect(p)
	SpawnLaunchEffect(p, 4*FRACUNIT/5, 0)
end

local function FloorBreak(mo, sec) --Function to break breakable floors from MAXControl by Inferno
	for x = mo.x, mo.x + mo.radius, mo.radius
		for y = mo.y, mo.y + mo.radius, mo.radius
			for fof in sec.ffloors()
				if (fof.flags & FF_EXISTS) and (fof.flags & FF_BUSTUP)
				and (mo.z + mo.height == fof.bottomheight or mo.z == fof.topheight
				or (mo.eflags & MFE_VERTICALFLIP) and (mo.z - mo.height == fof.topheight or mo.z == fof.bottomheight))
					EV_CrumbleChain(fof)
				end
			end
		end
	end
end

local function WallBounceDust(mo) --Dust from Krabs's Rebound Dash
	for i = 0, 9
		local dust = P_SpawnMobjFromMobj(mo, 0, 0, 18*mo.scale, MT_SPINDUST)
		dust.momx = (2*mo.momx) + (P_RandomRange(-10, 10) * FRACUNIT)
		dust.momy = (2*mo.momy) + (P_RandomRange(-10, 10) * FRACUNIT)
		dust.momz = P_RandomRange(-3, 3) * FRACUNIT
		dust.scale = P_RandomRange(50, 125) * FRACUNIT/100
		dust.destscale = 0
		dust.state = S_BOSSEXPLODE
	end
end

local function FloorBounceDust(mo) --Dust from MAXControl by Inferno
	for i = 0, 8
		local fa = (i*(ANGLE_180/4))
		local dust = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
		dust.scale = mo.scale
		dust.momx = FixedMul(sin(fa), FixedMul(20*FRACUNIT, mo.scale))
		dust.momy = FixedMul(cos(fa), FixedMul(20*FRACUNIT, mo.scale))
	end
end

addHook("PlayerSpawn", function(p)
	if p.mo and p.mo.valid and p.mo.skin == SKIN
		p.chargeability = {
			energy = MAXENERGY,
			charging = false,
			charge = 0,
			chargetime = 0,
			chargemod = 0,
			storespd = 0,
			stomp = false,
			wallbounce = 0,
			bouncestrength = 0,
			floorbounce = false,
			exhaust = false
		}
	end
end)

addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end

	if not p.chargeability
		p.chargeability = {
			energy = MAXENERGY,
			charging = false,
			charge = 0,
			chargetime = 0,
			chargemod = 0,
			storespd = 0,
			stomp = false,
			wallbounce = 0,
			bouncestrength = 0,
			floorbounce = false,
			exhaust = false
		}
	end
	local chrg = p.chargeability

	if p.powers[pw_carry]
	or P_PlayerInPain(p)
		chrg.wallbounce = 0
		chrg.stomp = false
		chrg.floorbounce = false
		StopCharge(p)
		return
	end

	if P_IsObjectOnGround(mo)
		chrg.wallbounce = 0
		chrg.stomp = false
		if chrg.floorbounce
			local WATERFACTOR = FRACUNIT
			if mo.eflags & MFE_UNDERWATER
				WATERFACTOR = 3*$/5
			end
			local SUPERFACTOR = FRACUNIT
			if p.powers[pw_super]
				SUPERFACTOR = 6*$/5
			end
			P_SetObjectMomZ(mo, FixedMul(FixedMul(FixedMul(68*FRACUNIT/9, chrg.bouncestrength), WATERFACTOR), SUPERFACTOR), false)
			p.pflags = $ | PF_JUMPED

			S_StartSound(mo, sfx_mspogo)
			mo.state = S_PLAY_ROLL

			FloorBounceDust(mo)
			if chrg.bouncestrength == 9*FRACUNIT/4
				FloorBreak(mo, mo.subsector.sector)
			end
			chrg.floorbounce = false
		end
	end

	if chrg.floorbounce
	and ((mo.eflags & MFE_SPRUNG) or mo.state == S_PLAY_SPRING)
		P_SetObjectMomZ(mo, FixedMul(mo.momz/3, chrg.bouncestrength), true)
		chrg.stomp = false
		chrg.floorbounce = false
	end

	if chrg.stomp
		local WATERFACTOR = FRACUNIT
		if mo.eflags & MFE_UNDERWATER
			WATERFACTOR = 3*$/5
		end
		local SUPERFACTOR = FRACUNIT
		if p.powers[pw_super]
			SUPERFACTOR = 6*$/5
		end
		P_SetObjectMomZ(mo, FixedMul(FixedMul(-64*FRACUNIT, WATERFACTOR), SUPERFACTOR), false)

		if mo.spriteyscale < 3*FRACUNIT/2
			mo.spriteyscale = $+FRACUNIT/10
		end

		if mo.spritexscale > 9*FRACUNIT/10
			mo.spritexscale = $-FRACUNIT/10
		end
	else
		mo.spritexscale = FRACUNIT
		mo.spriteyscale = FRACUNIT
	end

	if chrg.energy == 0
		chrg.exhaust = true
	elseif chrg.energy == MAXENERGY
		chrg.exhaust = false
	end

	if not chrg.charging
		if chrg.exhaust
			chrg.energy = min($ + 5*FRACUNIT/8, MAXENERGY) --Energy recovery while exhausted
		elseif P_IsObjectOnGround(mo)
			chrg.energy = min($ + 5*FRACUNIT/2, MAXENERGY) --Energy recovery while grounded
		end
		if p.powers[pw_super]
			chrg.energy = min($ + 5*FRACUNIT/8, MAXENERGY) --Additional energy recovery while Super
		end
	else
		if chrg.charge > 0 and chrg.charge < MAXCHARGE/2
			SpawnChargeOrb(p, 4)
		elseif chrg.charge >= MAXCHARGE/2 and chrg.charge < MAXCHARGE
			SpawnChargeOrb(p, 1)
		end
		if not p.powers[pw_super]
			if not p.prevcolor
				p.prevcolor = mo.color
			end
			if leveltime % 4 < 2
				mo.color = p.JMSVFXCOLOR
			else
				mo.color = p.prevcolor
			end
		end
		if leveltime % 4 < 2
			mo.colorized = true
		else
			mo.colorized = false
		end
	end

	if mo.state == S_PLAY_CHARGEJUMP
	or mo.state == S_PLAY_CHARGETHOK
	or chrg.stomp
		P_SpawnGhostMobj(mo)
		P_SpawnGhostMobj(p.jmsbooster1)
		P_SpawnGhostMobj(p.jmsbooster2)
	end

	chrg.wallbounce = max($ - 1, 0)
end)

addHook("AbilitySpecial", function(p)
	if not (p.mo and p.mo.health and p.mo.valid and p.mo.skin == SKIN) return end
	if p.powers[pw_carry] return end
	local chrg = p.chargeability

	if not chrg.charging
	and not (p.pflags & (PF_SPINDOWN | PF_JUMPDOWN))
	and chrg.energy
	and not chrg.exhaust
	and not p.Homing
		S_StartSound(p.mo, sfx_cdfm44)
		p.pflags = $ | PF_NOJUMPDAMAGE & ~PF_SPINNING

		chrg.charging = true
		chrg.storespd = p.speed
		chrg.wallbounce = 0
		chrg.stomp = false
		chrg.floorbounce = false
	end
end)

addHook("PreThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end
		if p.powers[pw_carry] return end
		if not p.chargeability return end
		local chrg = p.chargeability

		if P_PlayerInPain(p)
		or mo.eflags & MFE_SPRUNG
			StopCharge(p)
		end

		if chrg.charging
		and not P_IsObjectOnGround(mo)

			if p.cmd.buttons & BT_JUMP
			and chrg.chargetime <= MAXCHARGETIME
				mo.momx = 8*$/9
				mo.momy = 8*$/9
				mo.momz = 2*$/5
				mo.state = S_PLAY_CHARGE

				chrg.chargetime = $ + 1

				chrg.charge = min(max($ + CHRGSPD, MINCHARGE), min(chrg.energy, MAXCHARGE))
				if (p.powers[pw_super] or p.powers[pw_sneakers]) --Increase charging by 50% while super or having Super Sneakers
				and leveltime % 2 == 0
					chrg.charge = min($ + CHRGSPD, min(chrg.energy, MAXCHARGE))
				end
				if p.boostmode.active --Increase charging by 25% while in dash mode
				and leveltime % 4 == 0
					chrg.charge = min($ + CHRGSPD, min(chrg.energy, MAXCHARGE))
				end

				if p.cmd.buttons & BT_SPIN
				and not (p.pflags & PF_SPINDOWN)
					p.pflags = $ | PF_SPINDOWN & ~PF_NOJUMPDAMAGE
					if p.menu.contents[1][3].value
						doChargeJump(p)
					else
						doChargeStomp(p)
					end
					StopCharge(p)
				end

			elseif not (p.cmd.buttons & BT_JUMP)
				p.pflags = $ & ~PF_NOJUMPDAMAGE
				if P_GetPlayerControlDirection(p) ~= 0
					doChargeThok(p)
					StopCharge(p)
				else
					if p.menu.contents[1][3].value
						doChargeStomp(p)
					else
						doChargeJump(p)
					end
					StopCharge(p)
				end

			elseif chrg.chargetime > MAXCHARGETIME
				mo.state = S_PLAY_FALL
				p.pflags = $ | PF_THOKKED
				StopCharge(p)
			end
		elseif chrg.charging
		and P_IsObjectOnGround(mo)
			StopCharge(p)
		end
	end
end)

addHook("MobjLineCollide", function(mo, line) --Line detection code from Krabs's Rebound Dash
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end
	local p = mo.player
	local chrg = p.chargeability
	if not chrg return end

	local side = P_PointOnLineSide(mo.x, mo.y, line)
	local sector = nil

	if side == 1
		sector = line.frontsector
	else
		sector = line.backsector
	end

	if sector == nil return end

	for rover in sector.ffloors()
		if not (rover.flags & FF_EXISTS) or not (rover.flags & FF_BLOCKPLAYER)
			continue
		end

		local topheight = rover.topheight
		local bottomheight = rover.bottomheight

		if (rover.t_slope)
			topheight = P_GetZAt(rover.t_slope, mo.x, mo.y)
		end
		if (rover.b_slope)
			bottomheight = P_GetZAt(rover.b_slope, mo.x, mo.y)
		end

		if (topheight < mo.z)
			continue
		end

		if (bottomheight > mo.z + mo.height)
			continue
		end

		if (rover.flags & FF_BUSTUP)
			chrg.bustsector = sector
			chrg.bustrover = rover
		end
		break
	end
end, MT_PLAYER)

addHook("MobjMoveBlocked", function(mo)
	if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end
	local p = mo.player
	local chrg = p.chargeability
	if not chrg return end

	if chrg.wallbounce
		P_BounceMove(mo) --Bounce off walls
		mo.state = S_PLAY_ROLL

		mo.momx = FixedMul(FixedMul($, 16*FRACUNIT/99), chrg.bouncestrength)
		mo.momy = FixedMul(FixedMul($, 16*FRACUNIT/99), chrg.bouncestrength)
		P_SetObjectMomZ(mo, FixedMul(48*FRACUNIT/9, chrg.bouncestrength), false)

		S_StartSound(mo, sfx_mspogo)
		WallBounceDust(mo)
		chrg.wallbounce = 0

		if chrg.bustsector and chrg.bustsector.valid
		and chrg.bustrover and chrg.bustrover.valid
			if chrg.bouncestrength == 9*FRACUNIT/4
				EV_CrumbleChain(chrg.bustsector, chrg.bustrover) --Break bustable FOFs
			end
			chrg.bustsector = nil
			chrg.bustrover = nil
		end
	end
end, MT_PLAYER)

addHook("MobjMoveCollide", function(mo, mobj) --Break saloon doors with fully charged thoks
    if not (mo.valid and mo.skin == SKIN) return end
	local p = mo.player
	local chrg = p.chargeability
	if not chrg return end

	if chrg.bouncestrength == 9*FRACUNIT/4
		if mobj.type == MT_SALOONDOORCENTER
			local bouncex = P_ReturnThrustX(mo, mo.angle + ANGLE_180, FixedMul(FixedHypot(mo.momx, mo.momy), 4*FRACUNIT/11))
			local bouncey = P_ReturnThrustY(mo, mo.angle + ANGLE_180, FixedMul(FixedHypot(mo.momx, mo.momy), 4*FRACUNIT/11))
			mo.momx = bouncex
			mo.momy = bouncey			
			P_SetObjectMomZ(mo, 12*FRACUNIT, false)
			S_StartSound(mo, sfx_mspogo)

			chrg.wallbounce = 0
			p.pflags = $ | PF_NOJUMPDAMAGE
			mo.state = S_PLAY_ROLL

			WallBounceDust(mo)

			--Debris code is modified ground bounce dust from MAXControl by Inferno 
			for i = 0, 8
				local fa = (i*(ANGLE_180/4))
				local debris = P_SpawnMobjFromMobj(mobj, 0, 0, mobj.info.height/2, MT_WOODDEBRIS)
				debris.momx = FixedMul(sin(fa), 14*FRACUNIT)
				debris.momy = FixedMul(cos(fa), 14*FRACUNIT)
			end
			S_StartSound(mobj, sfx_wbreak)

			P_KillMobj(mobj, mo, mo)
		elseif mobj.type == MT_SALOONDOOR
			P_KillMobj(mobj, mo, mo)
		end
	end
end, MT_PLAYER)

addHook("MobjDamage", function(t, i, s)
	if not (i and i.player and i.skin == SKIN) return end
	local chrg = i.player.chargeability
	if not chrg return end

	if t and t.valid
		if t.flags & MF_ENEMY|MF_BOSS
			chrg.energy = min($ + 75*FRACUNIT/4, MAXENERGY)
		end
	end
end)

hud.add(function(v, p) --Charge Abilities energy meter, adapted from Inuyasha's Boss HUD example
	if p.mo and p.mo.skin == SKIN
	and p.powers[pw_carry] ~= CR_NIGHTSMODE
		local chrg = p.chargeability
		if not chrg return end

		local posx
		if modeattacking
			posx = 54
		else
			posx = 72
		end
		local posy = 178
		local maxmeter = 53

		local color = p.mo.color
		if p.prevcolor
			color = p.prevcolor
		end
		if p.menu.contents[1][1].value == 1
			color = SKINCOLOR_BLUE
		elseif p.menu.contents[1][1].value == 2
			color = SKINCOLOR_ORANGE
		elseif p.menu.contents[1][1].value == 3
			color = SKINCOLOR_RED
		elseif p.menu.contents[1][1].value == 4
			color = SKINCOLOR_ROSY
		elseif p.menu.contents[1][1].value == 5
			color = SKINCOLOR_PURPLE
		elseif p.menu.contents[1][1].value == 6
			color = SKINCOLOR_YELLOW
		elseif p.menu.contents[1][1].value == 7
			color = SKINCOLOR_GREEN
		end
		local invcolor = skincolors[color].invcolor
		if p.powers[pw_super]
			invcolor = skins[p.mo.skin].prefoppositecolor
		end

		local energymeter = maxmeter * chrg.energy / MAXENERGY
		local chargemeter = min(maxmeter * chrg.charge / MAXENERGY, energymeter)
		local chargemeterstart = maxmeter - energymeter
		local maxcharge = max(maxmeter / 2 - chargemeterstart + 1, 0)

		v.draw(posx, posy, v.cachePatch("JMS_STRT"), V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_PERPLAYER)
		posx = $ + 8

		for i = 1, maxmeter
			if energymeter >= i
				v.draw(posx, posy, v.cachePatch("JMS_ON"), V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_PERPLAYER, v.getColormap(TC_DEFAULT, color))
			else
				v.draw(posx, posy, v.cachePatch("JMS_OFF"), V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_PERPLAYER)
			end
			if maxcharge == i
				v.draw(posx, posy, v.cachePatch("JMS_ON"), V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_PERPLAYER, v.getColormap(TC_DEFAULT, invcolor))
			end
			posx = $ + 1
		end

		v.draw(posx, posy, v.cachePatch("JMS_END"), V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_PERPLAYER)

		local posX = (posx - 20)*FRACUNIT
		local posY = (posy + 2)*FRACUNIT + FRACUNIT/2

		if chrg.exhaust
			local alpha = max(9 - leveltime % 15, 0) * V_10TRANS
			v.drawString(posX, posY, "EXHAUSTED", V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_PERPLAYER|alpha|V_REDMAP, "thin-fixed-center")
		end

		posx = $ - chargemeterstart - 1
		for i = 1, maxmeter
			if chargemeter >= i
				v.draw(posx, posy, v.cachePatch("JMS_ON"), V_50TRANS|V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_PERPLAYER, v.getColormap(TC_DEFAULT, invcolor))
			end
			posx = $ - 1
		end
	end
end, "game")