--Dash mode
local SKIN = "jetmetalsonic"

freeslot("S_PLAY_DASHREADY")

states[S_PLAY_DASHREADY] = {SPR_PLAY, SPR2_CHRG|A, 1, nil, 0, 0, S_PLAY_DASHREADY}

local MAXCHARGE = 3*TICRATE		--Charge required to activate Dash Mode
local CHARGEDECAY = 3*TICRATE/7	--Amount of charge lost per tic
local SPINDASHSPD = 90*FRACUNIT	--Immediate spindash charge in Dash Mode

addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == SKIN) return end

	if not p.boostmode --Create boostmode userdata structure
		p.boostmode = {
			charge = 0,
			chargespd = p.runspeed,
			chargespd2 = p.runspeed + abs(p.runspeed - skins[mo.skin].normalspeed) / 2,
			active = false,
			peelout = false,
			speedcap = 3*skins[mo.skin].normalspeed/2
		}
	end
	local dash = p.boostmode

	if p.playerstate == PST_DEAD
	or p.exiting
	or p.powers[pw_carry]
		dash.charge = 0
	end

	if p.speed >= dash.chargespd
	or (p.pflags & PF_STARTDASH)
		dash.charge = min($ + 1, MAXCHARGE) --Increase charge if moving at chargespd or charging a spindash

		if (p.speed >= dash.chargespd2 or (p.pflags & PF_STARTDASH))
		and leveltime % 2
			dash.charge = min($ + 1, MAXCHARGE) --Increase charge 50% faster when moving at chargespd2 or charging a spindash
		end

	elseif P_IsObjectOnGround(mo)
	or (not P_IsObjectOnGround(mo) and not dash.active)
		dash.charge = max($ - CHARGEDECAY, 0) --Decrease charge if on the ground (or in the air without Dash Mode active) and not moving fast enough
	end

	if dash.charge > 7*MAXCHARGE/10 and dash.charge < MAXCHARGE
	and not dash.active
		SpawnChargeOrb(p, 2)
	end

	if dash.charge == MAXCHARGE
	and not dash.active
		S_StartSound(mo, sfx_kc4d) --sfx_cdfm40
		if not (p.pflags & PF_STARTDASH)
			P_StartQuake(20*FRACUNIT, TICRATE/5)
			SpawnLaunchEffect(p, FRACUNIT, 1)
		end

		p.charflags = $ | SF_RUNONWATER
		p.jumpfactor = 11*$/10
		p.mindash = SPINDASHSPD
		p.maxdash = SPINDASHSPD
		p.spinitem = MT_NULL

		dash.active = true

	elseif not dash.charge
	and dash.active
		S_StartSound(mo, sfx_kc65)

		p.charflags = $ & ~SF_RUNONWATER
		p.charability2 = CA2_SPINDASH
		p.jumpfactor = skins[mo.skin].jumpfactor
		p.mindash = 15*FRACUNIT
		p.maxdash = 70*FRACUNIT
		p.spinitem = MT_THOK

		dash.active = false
	end
	if p.powers[pw_carry] return end

	--Automatic uncurl when in Dash Mode, based on code by Frostiikin
	if (p.pflags & PF_STARTDASH)
		if dash.active
			mo.state = S_PLAY_DASHREADY
			dash.peelout = true
		end
	end
	if mo.state == S_PLAY_ROLL
	and P_IsObjectOnGround(mo)
	and dash.peelout
		mo.state = S_PLAY_RUN
		p.pflags = $ & ~PF_SPINNING
		SpawnLaunchEffect(p, FRACUNIT, 1)
		dash.peelout = false
	end

	if dash.active
		if leveltime % 6 == 0
			local ghost = P_SpawnGhostMobj(mo)
			ghost.tracer = mo
			ghost.colorized = true
			ghost.blendmode = AST_ADD
			ghost.scale = mo.scale
			ghost.destscale = 4*mo.scale/3
			ghost.scalespeed = abs(ghost.scale - ghost.destscale) / ghost.fuse
			ghost.dashactive = true
		end

		if P_IsObjectOnGround(mo)
		and not (mo.eflags & MFE_JUSTHITFLOOR)
			p.charability2 = CA2_SPINDASH
			if not (p.pflags & PF_STARTDASH)
				p.charflags = $ | SF_RUNONWATER
			end
		else
			p.charability2 = CA2_NONE
		end

		if p.pflags & PF_SPINNING
		and not P_IsObjectOnGround(mo)
			mo.state = S_PLAY_ROLL
			p.pflags = $ & ~PF_SPINNING
		end

		if p.speed >= p.runspeed
			p.normalspeed = min($ + skins[mo.skin].normalspeed / 500, dash.speedcap)

			if mo.state == S_PLAY_RUN
				mo.state = S_PLAY_DASH
				SpawnAuraEffect(p)
			end

		elseif mo.state == S_PLAY_DASH
			mo.state = S_PLAY_RUN
		end

	else
		p.normalspeed = skins[mo.skin].normalspeed

		if mo.state == S_PLAY_DASH
		and not p.Homing
			mo.state = S_PLAY_RUN
		end

		dash.peelout = false
	end
end)

addHook("SpinSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == SKIN) return end
	if p.powers[pw_carry] return end

	if p.speed > FixedMul(5*FRACUNIT, mo.scale)
	and p.pflags & PF_STARTDASH
		for i = 0, 3
			P_SpawnSkidDust(p, FixedMul(20*FRACUNIT, mo.scale/2))
		end
		return true
	end

	if p.boostmode.active
	and P_IsObjectOnGround(mo)
	and mo.state ~= S_PLAY_SPINDASH
	and not (p.pflags & PF_SPINNING)
		p.charflags = $ & ~SF_RUNONWATER
		p.dashspeed = p.mindash
		mo.state = S_PLAY_SPINDASH
		p.pflags = $|PF_USEDOWN|PF_STARTDASH|PF_SPINNING
		S_StartSound(mo, sfx_spndsh)
		return true
	end
end)

addHook("PlayerCanDamage", function(p)
	if not (p.mo and p.mo.valid and p.mo.skin == SKIN) return end
	if p.boostmode.active
		return true
	end
end)

addHook("MobjMoveCollide", function(mo, mobj) --Spike breaking adapted from code by ARJr.
    if not (mo and mo.valid and mo.skin == SKIN) return end

	if mo.player.boostmode.active
    and mobj.type == MT_SPIKE
		if (mobj.z >= (mo.z - mobj.height/2)) and (mobj.z <= (mo.z + mo.height + mobj.height/2))
			P_KillMobj(mobj, mo, mo)
		elseif (mobj.flags2 & MF2_OBJECTFLIP)
		and (mobj.z <= (mo.z - mobj.height/2)) and (mobj.z >= (mo.z + mo.height+ mobj.height/2))
			P_KillMobj(mobj, mo, mo)
        end
    end
end, MT_PLAYER)

addHook("MobjThinker", function(mo)
	if not (mo.tracer and mo.tracer.valid) return end
	local t = mo.tracer
	local p = t.player

	if mo.dashactive
		local z_adjust = 0
		if t.eflags & MFE_VERTICALFLIP
			z_adjust = FixedMul(t.scale - mo.scale, P_GetPlayerHeight(p))
			if t.state == S_PLAY_ROLL
			or t.state == S_PLAY_DASHREADY
				z_adjust = $ - P_GetPlayerHeight(p)/3
			end
		end
		if p.prevcolor
			mo.color = p.prevcolor
		else
			mo.color = t.color
		end
		P_TeleportMove(mo, t.x, t.y, t.z + z_adjust)
	end
end, MT_GHOST)