--Homing thok
local SKIN = "jetmetalsonic"

local function CustomLookForEnemies(p, ringdist, span, above, disregard) --Enemy search. Code re-used and modified from CobaltBW's Fang Popgun Modifications
	local twod = (twodlevel or p.mo.flags2 & MF2_TWOD)
	if (twod)
		ringdist = $/2
		span = 2*$/3
	end

	local maxdist = FixedMul(ringdist, p.mo.scale)
	local closestdist = 0
	local closestmo = nil
	local nonenemiesdisregard = disregard
	searchBlockmap("objects", function(pmo, mo)
		if (mo.flags & MF_NOCLIPTHING) return end
		if (mo.health <= 0) return end

		if not(mo.p)
		and (!((mo.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR)
		and (mo.flags & MF_SHOOTABLE)) or (mo.flags & MF_SPRING)) == !(mo.flags2 & MF2_INVERTAIMABLE))
			return
		end
		if (mo == pmo) return end
		if (mo.flags2 & MF2_FRET) return end
		if (mo.flags & nonenemiesdisregard) return end
		if mo.type == MT_DETON or mo.type == MT_TNTBARREL or mo.type == MT_METALSONIC_BATTLE return end

		//Disallow targeting teammates
		if mo.p and (mo.p.spectator or gametyperules&GTR_FRIENDLY or G_GametypeHasTeams() and mo.p.ctfteam == p.ctfteam)
			return
		end

		//Do angle/distance checks
		local zdist = (pmo.z + pmo.height/2) - (mo.z + mo.height/2)
		local dist = P_AproxDistance(pmo.x - mo.x, pmo.y - mo.y)
		//CBW:	Made the angle checks their own locals, for readability purposes
		//		I also unsigned the angle checks, which appears to correct failed OutOfBounds checks above the player.
		local xyz_angle = abs(R_PointToAngle2(0, 0, dist, zdist))
		local xy_angle = abs(R_PointToAngle2(
				pmo.x + P_ReturnThrustX(pmo, pmo.angle, pmo.radius),
				pmo.y + P_ReturnThrustY(pmo, pmo.angle, pmo.radius),
				mo.x, mo.y) - pmo.angle)

		dist = P_AproxDistance(dist, zdist)
		if (dist > maxdist)
			return -- out of range
		end
		if (xyz_angle > span)
			return -- Don't home outside of desired angle!
		end

		if not above -- Don't home above the player if not desired!
			if pmo.eflags & MFE_VERTICALFLIP
			and zdist > 0
				return
			elseif zdist < 0
				return
			end
		end

		if (twod
		and abs(pmo.y-mo.y) > pmo.radius)
			return -- not in your 2d plane
		end

		if ((closestmo and closestmo.valid) and (dist > closestdist))
			return
		end
		if (xy_angle > span)
			return -- behind back
		end

		if not (P_CheckSight(pmo, mo))
			return -- out of sight
		end

		closestmo = mo
		closestdist = dist
	end, p.mo, p.mo.x-maxdist, p.mo.x+maxdist, p.mo.y-maxdist, p.mo.y+maxdist)
	return closestmo
end

local function CustomHoming(mo, t, spdmod)
	mo.angle = R_PointToAngle2(mo.x, mo.y, t.x, t.y)

	local dist = P_AproxDistance(P_AproxDistance(t.x - mo.x, t.y - mo.y), t.z - mo.z)
	local speed = FixedMul(FixedMul(mo.player.actionspd, spdmod), mo.scale)

	local waterfactor = FRACUNIT
	if mo.eflags & MFE_UNDERWATER
		waterfactor = 3*$/5
	end
	speed = FixedMul($, waterfactor)

	mo.momx = FixedMul(FixedDiv(t.x - mo.x, dist), speed)
	mo.momy = FixedMul(FixedDiv(t.y - mo.y, dist), speed)
	mo.momz = FixedMul(FixedDiv(t.z - mo.z, dist), speed)
end

addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.health and mo.skin == SKIN) return end
	if p.powers[pw_carry] return end

	if not p.Homing
	and not (p.pflags & PF_THOKKED)
	and not P_IsObjectOnGround(mo)
	and not P_PlayerInPain(p)
		mo.tracer = CustomLookForEnemies(p, 6*RING_DIST/5, ANGLE_45, true, MF_SPRING|MF_MONITOR)
		if mo.tracer and mo.tracer.valid and mo.tracer.health
		and not (mo.state == S_PLAY_SPRING or mo.state == S_PLAY_FALL)
		and not p.charging
			P_SpawnLockOn(p, mo.tracer, S_LOCKON2)
		end
	end

	if P_PlayerInPain(p)
	or (mo.eflags & MFE_SPRUNG)
		p.Homing = 0
	end
	if p.Homing
		p.powers[pw_nocontrol] = 1

		mo.state = S_PLAY_DASH

		P_SpawnGhostMobj(mo)
		P_SpawnGhostMobj(p.jmsbooster1)
		P_SpawnGhostMobj(p.jmsbooster2)

		if not (mo.tracer and mo.tracer.valid and mo.tracer.health)
			p.Homing = 0
		end

		if not p.Homing
			mo.state = S_PLAY_FALL
			p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
		else
			p.pflags = $ & ~PF_NOJUMPDAMAGE
			CustomHoming(mo, mo.tracer, 2*FRACUNIT/3)
			p.Homing = $ - 1
		end
	end
end)

addHook("JumpSpinSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end
	if p.powers[pw_carry] return end
	local chrg = p.chargeability

	if not p.Homing
	and not (p.pflags & (PF_SPINDOWN|PF_THOKKED))
	and mo.tracer and mo.tracer.valid and mo.tracer.health
		S_StartSound(mo, sfx_beflap)
		mo.angle = R_PointToAngle2(mo.x, mo.y, mo.tracer.x, mo.tracer.y)
		SpawnAuraEffect(p)
		SpawnLaunchEffect(p, 4*FRACUNIT/5, 1)

		p.Homing = 3*TICRATE
		p.pflags = $ & ~PF_NOJUMPDAMAGE

		chrg.wallbounce = 0
		chrg.stomp = false
		chrg.floorbounce = false
	end
end)

addHook("MobjMoveBlocked", function(mo)
	if not (mo and mo.health and mo.valid and mo.skin == SKIN) return end
	local p = mo.player

	if p.Homing
		p.Homing = 0
		mo.state = S_PLAY_FALL
		p.pflags = $|PF_NOJUMPDAMAGE|PF_THOKKED
	end
end, MT_PLAYER)

addHook("MobjDamage", function(t, i, s)
	if not (i and i.player and i.skin == SKIN) return end
	local p = i.player

	if t.valid
	and (t.flags & MF_ENEMY|MF_BOSS|MF_MONITOR)
	and p.Homing
		local speed = p.actionspd
		if i.eflags & MFE_UNDERWATER
			speed = 3*$/5
		end
		P_InstaThrust(i, i.angle, FixedMul(speed/6, i.scale))
		P_SetObjectMomZ(i, speed/10, false)
		p.pflags = $ & ~PF_THOKKED
		i.state = S_PLAY_ROLL
		p.Homing = 0
	end
end)