--Miscellaneous
local SKIN = "jetmetalsonic"

freeslot("S_PLAY_DEAD_ON_FLOOR")

states[S_PLAY_DEAD_ON_FLOOR] = {SPR_PLAY, SPR2_CNT1|FF_ANIMATE, -1, nil, 5, 4, S_PLAY_DEAD_ON_FLOOR}

local function SpawnSuperEffect(p)
	for i = 1, 8
		local fa = (i*(ANGLE_180/4))
		local radius = FixedMul(states[S_JMSSUPEREFF].var1 / states[S_JMSSUPEREFF].var2, p.mo.scale)

		local offsetx = FixedMul(cos(fa), radius)
		local offsety = FixedMul(sin(fa), radius)

		local effect = P_SpawnMobjFromMobj(p.mo, offsetx, offsety, p.mo.height/2, MT_JMSSUPEREFF)
		effect.scale = p.mo.scale
		effect.color = p.JMSVFXCOLOR
		effect.angle = fa + ANGLE_90
	end
end

--Press jump or spin while springing to recurl
addHook("SpinSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == SKIN) return end
	if p.powers[pw_carry] return end
	if p.canrecurl
	and not (p.pflags & PF_SPINDOWN)
		p.canrecurl = false
		p.pflags = $|PF_JUMPED|PF_SPINDOWN
		mo.state = S_PLAY_ROLL
		S_StartSound(mo, sfx_3db16)
		return true
	end
end)
addHook("JumpSpecial", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == SKIN) return end
	if p.powers[pw_carry] return end
	if p.canrecurl
	and not (p.pflags & PF_JUMPDOWN)
		p.canrecurl = false
		p.pflags = $|PF_JUMPED|PF_JUMPDOWN
		mo.state = S_PLAY_ROLL
		S_StartSound(mo, sfx_3db16)
		return true
	end
	if mo.state == S_PLAY_DASHREADY
		mo.state = S_PLAY_ROLL
	end
end)

addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid and mo.skin == SKIN) return end
	local chrg = p.chargeability
	if not chrg return end
	if p.powers[pw_carry] return end

	--No air walking/standing
	if p.pflags & ~(PF_JUMPED|PF_SPINNING)
	and not P_PlayerInPain(p)
	and not P_IsObjectOnGround(mo)
	and not p.powers[pw_nocontrol]
		if mo.state == S_PLAY_STND
		or mo.state == S_PLAY_WAIT
		or mo.state == S_PLAY_WALK
		or mo.state == S_PLAY_RUN
		or mo.state == S_PLAY_DASH
			mo.state = S_PLAY_SPRING
		end
	end

	if mo.eflags & MFE_SPRUNG
		p.canrecurl = true
	elseif P_IsObjectOnGround(mo)
	or P_PlayerInPain(p)
	or p.playerstate == PST_DEAD
		p.canrecurl = false
	end

	if p.playerstate == PST_DEAD --Custom death animation
		chrg.wallbounce = 0
		chrg.stomp = false
		chrg.floorbounce = false

		mo.flags = $ & ~MF_NOCLIPHEIGHT

		if P_IsObjectOnGround(mo)
			mo.state = S_PLAY_DEAD_ON_FLOOR
			if p.deadalready
				mo.momx = 0
				mo.momy = 0
				mo.rollangle = 0
			end
		else
			mo.state = S_PLAY_DEAD
			mo.rollangle = $ + ANG10
		end

		if not p.deadalready
			P_InstaThrust(mo, mo.angle, p.prevspeed/2)
			mo.momz = 2*$/3
			p.deadalready = true
		end
	elseif p.deadalready
		p.deadalready = false
	end
	p.prevspeed = FixedHypot(mo.momx, mo.momy)

	--Rebind Super transformation
	local superbutton = BT_CUSTOM1
	if p.menu.contents[1][4].value == 1
		superbutton = BT_CUSTOM2
	elseif p.menu.contents[1][4].value == 2
		superbutton = BT_CUSTOM3
	elseif p.menu.contents[1][4].value == 3
		superbutton = BT_TOSSFLAG
	end
	if All7Emeralds(emeralds) and (p.rings > 49) and (p.playerstate == PST_LIVE)
	and not p.powers[pw_super]
	and not (p.pflags & PF_SPINNING) and not (p.pflags & PF_THOKKED)
	and not (p.panim == PA_ABILITY) and not (p.panim == PA_ABILITY2) and not (p.panim == PA_PAIN) and not (p.panim == PA_ETC)
	and not (mo.state == S_PLAY_SPRING) and not (mo.state == S_PLAY_FALL)
	and not chrg.stomp and not chrg.charging and not p.Homing
	and not P_IsObjectOnGround(mo)
		if p.cmd.buttons & superbutton
			p.charflags = $ | SF_SUPER
			P_DoSuperTransformation(p)
			P_StartQuake(20*FRACUNIT, TICRATE/2)
			SpawnSuperEffect(p)
		end
	elseif not p.powers[pw_super]
		p.charflags = $ & ~SF_SUPER
	end
end)

--Combo counter
addHook("PreThinkFrame"	, do
	for p in players.iterate
		if not (p.mo and p.mo.valid and p.mo.skin == SKIN) return end
		if p.powers[pw_carry] return end
		if not p.combocount
			p.combocount = 0
		end
		if p.mo.eflags & MFE_JUSTHITFLOOR
		and not p.chargeability.floorbounce
			p.combocount = 0
		end
	end
end)

addHook("MobjDamage", function(t, i, s)
	if not (i and i.player and i.skin == SKIN) return end
	if not P_IsObjectOnGround(i)
	and t.valid
		if t.flags & MF_ENEMY|MF_BOSS
			i.player.combocount = $ + 1
		end
	end
end)

hud.add(function(v, p) --Combo counter on the HUD
	if p.mo and p.mo.skin == SKIN
	and p.powers[pw_carry] ~= CR_NIGHTSMODE
		if p.combocount > 1
			local posx = 41
			local posy = 152
			if modeattacking
				posy = 141
			end

			--Counter cannot display beyond 2 digits
			if p.combocount > 99
				v.drawPaddedNum(posx, posy, 99, 2, V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
			else
				v.drawPaddedNum(posx, posy, p.combocount, 2, V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
			end

			posx = $ - 25
			posy = $ + 12
			v.drawString(posx, posy, "hits!", V_PERPLAYER|V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_YELLOWMAP, "left")
		end
	end
end, "game")